{% if properties %}
<span class="doc doc-properties">
  {% for property in properties %} {% if property != "pydantic-model" %}
  <small class="doc doc-property doc-property-{{ property }}">
    <code>{{ property }}</code>
  </small>
  {% endif %} {% endfor %}
</span>
{% endif %}
